QTextEdit#html,
QPlainTextEdit#plain {
  font: 8pt Liberation Mono;
  letter-spacing: 0px;
}

QPushButton#topbar-status {
  border: none;
  border-bottom: 1px solid black;
  padding-left: 9px;
  padding-top: 1px;
  margin-top: 1px;
  qproperty-iconSize: 13px;
}

QLabel#topbar-title {
  font: 8pt Liberation Mono;
  border-bottom: 1px solid black;
  padding: 4px;
  margin-top: 1px;
  qproperty-alignment: 'AlignCenter';
  qproperty-wordWrap: true;
}

QPushButton#topbar-close {
  border: none;
  border-bottom: 1px solid black;
  font: 11pt Liberation Mono;
  font-weight: bold;
  padding-right: 9px;
  padding-top: 1px;
  min-height: 12px;
  margin-top: 1px;
}

QLabel#msgbar-folder,
QLabel#msgbar-popup-left {
  font: 6pt Liberation Mono;
  padding-left: 2px;
  qproperty-alignment: 'AlignLeft | AlignBottom';
}

QLabel#msgbar-pixels,
QLabel#msgbar-popup-right,
QLabel#msgbar-words {
  font: 6pt Liberation Mono;
  padding-right: 2px;
  qproperty-alignment: 'AlignRight | AlignBottom';
}

QPushButton#toolbar-icon {
  min-width: 0px;  /* Required to avoid oversized width */
}

QFrame#toolbar-frame {
  border: none;
}

QScrollBar:horizontal {
  height: 11px;
}

QScrollBar:vertical {
  width: 12px;
}

QScrollBar::add-line,
QScrollBar::sub-line,
QScrollBar::add-page:horizontal,
QScrollBar::sub-page:horizontal,
QScrollBar::add-page:vertical,
QScrollBar::sub-page:vertical
{
  background: none;
  border: none;
}

QWidget#lockscreen-frame {
  border-radius: 3px;
}

QLineEdit#lockscreen-input {
  border-radius: 2px;
}

QLabel#lockscreen-label {
  padding: 8px;
}
