NoteDecorations {
  msgbar-visible: true;
  popup-visible: true;
  toolbar-visible: false;
}

QSizeGrip,
QFrame#hotbar-frame,
QFrame#hotbar-spacer {
  height: 7px;
}

QSizeGrip#left,
QSizeGrip#right {
  width: 40px;
}

QSizeGrip#center {
  width: 70px;
}
