NoteDecorations {
  preview-fg: #d5d5d5;
  preview-bg: #2b2b2b;
  highlight-fg: #eeeeee;
  highlight-bg: #5e5e5e;
  encryption-icon: #7f7f7f;
  encryption-icon-active: #d5d5d5;
  toolbar-icon: #7f7f7f;
  toolbar-icon-active: #d5d5d5;
  status-icon: #7f7f7f;
  status-icon-active: #d5d5d5;
  border-color: #000000;
}

QTextEdit#html,
QPlainTextEdit#plain {
  color: #7f7f7f;
  background-color: #2b2b2b;
}

QTextEdit:active#html,
QPlainTextEdit:active#plain {
  color: #c5c5c5;
}

QTextEdit[readOnly="true"]#html,
QPlainTextEdit[readOnly="true"]#plain {
  color: #7f7f7f;
}

QFrame#html,
QFrame#plain {
  border: 4px solid #2b2b2b;
}

AbstractText,
QSizeGrip,
Vim,
QFrame#hotbar-frame,
QFrame#hotbar-spacer {
  color: #7f7f7f;
  background-color: #262626;
}

QLabel#topbar-title,
QPushButton#topbar-status,
QPushButton#topbar-close {
  color: #7f7f7f;
  background-color: #262626;
  border: none;
}

QLabel:active#topbar-title,
QPushButton:active#topbar-close {
  color: #d5d5d5;
}

QPushButton:hover#topbar-close {
  color: #ffffff;
}

QFrame#popup-frame,
QFrame#msgbar-frame,
QLabel#msgbar-folder,
QLabel#msgbar-pixels,
QLabel#msgbar-popup-left,
QLabel#msgbar-popup-right,
QLabel#msgbar-words {
  color: #7f7f7f;
  background-color: #2b2b2b;
}

QFrame:active#msgbar-frame,
QLabel:active#msgbar-folder,
QLabel:active#msgbar-pixels,
QLabel:active#msgbar-popup-left,
QLabel:active#msgbar-popup-right,
QLabel:active#msgbar-words {
  color: #a0a0a0;
}

QFrame#toolbar-frame,
QLabel#toolbar-spacer,
QLabel#toolbar-separator {
  color: #262626;
  background-color: #262626;
  border: none;
}

QPushButton:hover#toolbar-icon {
  background-color: #5e5e5e;
}

QPushButton#toolbar-icon,
QPushButton:pressed#toolbar-icon {
  background-color: #262626;
}

QSizeGrip:hover {
  background-color: #5e5e5e;
}

QScrollBar {
  background: #5e5e5e;
}

QScrollBar::handle {
  background-color: #404040;
}

QWidget#lockscreen {
  background-color: #262626;
}

QWidget#lockscreen-frame {
  background-color: #5e5e5e;
}

QLabel#lockscreen-label {
  color: #7f7f7f;
}

QLabel:active#lockscreen-label {
  color: #eeeeee;
}

QLineEdit#lockscreen-input {
  color: #696969;
  background-color: #ffffff;
}

QLineEdit:active#lockscreen-input {
  color: #000000;
}

QLineEdit[error="true"]#lockscreen-input {
  color: #9c0006;
  background-color: #ffc7ce;
}
