NoteDecorations {
  msgbar-visible: false;
  popup-visible: true;
  toolbar-visible: true;
}

QSizeGrip,
QFrame#hotbar-frame,
QFrame#hotbar-spacer {
  height: 3px;
}

QSizeGrip#left,
QSizeGrip#right {
  width: 35px;
}

QSizeGrip#center {
  width: 70px;
}

QFrame#toolbar-frame {
  padding-top: 2px;
  border-top: 1px solid black;
}

QLabel#toolbar-spacer {
  border: none;
}

QLabel#toolbar-separator {
  font: 6pt Liberation Mono;
  font-weight: bold;
  padding-bottom: 1px;
}

QPushButton#toolbar-icon {
  border: none;
  padding: 4px;
  min-width: 16px;
  qproperty-iconSize: 13px;
}
